import aiohttp
import threading

from kiroframe_arcee.platform import CollectorFactory
from kiroframe_arcee.collectors.command_line import (
    Collector as CommandCollector)
from kiroframe_arcee.collectors.git import Collector as GitCollector
from kiroframe_arcee.collectors.hardware import Collector as HardwareCollector
from kiroframe_arcee.collectors.module import Collector as ImportsCollector
from kiroframe_arcee.collectors.console import Collector as OutCollector


def check_shutdown_flag_set(function):
    async def inner(self, *args, **kwargs):
        if not self.shutdown_flag.is_set():
            return await function(self, *args, **kwargs)

    return inner


class Sender:
    # default Kiroframe url
    base_url = "https://my.kiroframe.com:443/arcee/v2"

    def __init__(self, endpoint_url=None, ssl=True, shutdown_flag=None):
        if endpoint_url is None:
            endpoint_url = self.base_url
        self.endpoint_url = endpoint_url
        self.shutdown_flag = shutdown_flag or threading.Event()
        self.ssl = ssl

    @staticmethod
    async def m():
        platform = await CollectorFactory.get()
        return await platform().get_platform_meta()

    @staticmethod
    async def _proc_data():
        return await HardwareCollector.collect_stats()

    @staticmethod
    async def _imports_data():
        return await ImportsCollector.get_imports()

    @staticmethod
    async def _git_data():
        return await GitCollector.collect()

    @staticmethod
    async def _self_command():
        return await CommandCollector.collect()

    @staticmethod
    async def _output():
        return await OutCollector.collect()

    async def send_get_request(self, url, headers=None, params=None) -> dict:
        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.get(
                url, params=params, raise_for_status=True, ssl=self.ssl
            ) as response:
                return await response.json()

    async def send_post_request(self, url, headers=None, data=None) -> dict:
        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.post(
                url, json=data, raise_for_status=True, ssl=self.ssl
            ) as response:
                return await response.json()

    async def send_patch_request(self, url, headers=None, data=None) -> dict:
        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.patch(
                url, json=data, raise_for_status=True, ssl=self.ssl
            ) as response:
                return await response.json()

    @check_shutdown_flag_set
    async def get_run_id(self, task_key, token, run_name):
        uri = "%s/tasks/%s/run" % (self.endpoint_url, task_key)
        headers = {"x-api-key": token, "Content-Type": "application/json"}
        data = {
            "imports": await self._imports_data(),
            "git": await self._git_data(),
            "command": await self._self_command(),
            "name": run_name
        }
        return await self.send_post_request(uri, headers, data)

    @check_shutdown_flag_set
    async def add_milestone(self, run_id, token, value):
        uri = "%s/run/%s/milestones" % (self.endpoint_url, run_id)
        headers = {"x-api-key": token, "Content-Type": "application/json"}
        return await self.send_post_request(uri, headers, {"milestone": value})

    @check_shutdown_flag_set
    async def add_tags(self, run_id, token, tags):
        uri = "%s/run/%s" % (self.endpoint_url, run_id)
        headers = {"x-api-key": token, "Content-Type": "application/json"}
        return await self.send_patch_request(uri, headers, {"tags": tags})

    @check_shutdown_flag_set
    async def change_state(self, run_id, token, state, finish=False):
        uri = "%s/run/%s" % (self.endpoint_url, run_id)
        headers = {"x-api-key": token, "Content-Type": "application/json"}
        return await self.send_patch_request(
            uri, headers, {"state": state, "finish": finish}
        )

    @check_shutdown_flag_set
    async def create_stage(self, run_id, token, name):
        uri = "%s/run/%s/stages" % (self.endpoint_url, run_id)
        headers = {"x-api-key": token, "Content-Type": "application/json"}
        return await self.send_post_request(uri, headers, {"stage": name})

    @check_shutdown_flag_set
    async def send_stats(self, token, data):
        headers = {"x-api-key": token, "Content-Type": "application/json"}
        meta = await self.m()
        data.update({"platform": meta.to_dict()})
        await self.send_post_request(
            "%s/%s" % (self.endpoint_url, "collect"), headers, data
        )

    @check_shutdown_flag_set
    async def send_proc_data(self, run_id, token):
        uri = "%s/run/%s/proc" % (self.endpoint_url, run_id)
        headers = {"x-api-key": token, "Content-Type": "application/json"}
        data = dict()
        meta = await self.m()
        proc = await self._proc_data()
        data.update({"platform": meta.to_dict()})
        data.update({"proc_stats": proc})
        return await self.send_post_request(uri, headers, data)

    @staticmethod
    def generate_description(task_key, run_name, run_id):
        return f"Discovered in training {task_key} - {run_name}({run_id})"

    @check_shutdown_flag_set
    async def register_dataset(self, token, run_id, run_name, task_key, path,
                               dataset_name=None, description=None,
                               labels=None):
        uri = f"{self.endpoint_url}/run/{run_id}/dataset_register"
        headers = {"x-api-key": token, "Content-Type": "application/json"}

        if dataset_name is None:
            dataset_name = path
        if not description:
            description = self.generate_description(task_key, run_name, run_id)
        if labels is not None and not isinstance(labels, list):
            labels = [labels]

        data = {
            "path": path,
            "name": dataset_name,
            "description": description,
            "labels": labels or []
        }
        await self.send_post_request(uri, headers, data)

    @check_shutdown_flag_set
    async def add_hyperparams(self, run_id, token, hyperparams):
        uri = "%s/run/%s" % (self.endpoint_url, run_id)
        headers = {"x-api-key": token, "Content-Type": "application/json"}
        return await self.send_patch_request(
            uri, headers, {"hyperparameters": hyperparams}
        )

    async def send_console(self, run_id, token):
        uri = f"{self.endpoint_url}/run/{run_id}/consoles"
        headers = {"x-api-key": token, "Content-Type": "application/json"}

        data = await self._output()
        await self.send_post_request(uri, headers, data)

    @check_shutdown_flag_set
    async def add_model(self, token, key):
        headers = {"x-api-key": token, "Content-Type": "application/json"}
        model = await self.send_post_request(
            self.endpoint_url + '/models', headers, {"key": key}
        )
        return model.get('_id')

    @check_shutdown_flag_set
    async def create_model_version(self, run_id, model_id, token, path=None):
        headers = {"x-api-key": token, "Content-Type": "application/json"}
        uri = f'{self.endpoint_url}/runs/{run_id}/models/{model_id}/version'
        body = {'path': path} if path else {}
        await self.send_post_request(uri, headers, body)

    @check_shutdown_flag_set
    async def patch_model_version(self, run_id, model_id, token, params):
        headers = {"x-api-key": token, "Content-Type": "application/json"}
        uri = f'{self.endpoint_url}/runs/{run_id}/models/{model_id}/version'
        await self.send_patch_request(uri, headers, params)

    async def add_version(self, run_id, model_id, token, version):
        body = {'version': str(version)}
        await self.patch_model_version(run_id, model_id, token, body)

    async def add_version_aliases(self, run_id, model_id, token, aliases):
        body = {'aliases': aliases}
        await self.patch_model_version(run_id, model_id, token, body)

    async def add_version_tags(self, run_id, model_id, token, tags):
        body = {'tags': tags}
        await self.patch_model_version(run_id, model_id, token, body)

    @check_shutdown_flag_set
    async def add_artifact(self, token, run_id, run_name, task_key, path,
                           name=None, description=None, tags=None):
        headers = {"x-api-key": token, "Content-Type": "application/json"}
        body = {
            'run_id': run_id,
            'path': path
        }
        if name:
            body['name'] = name
        if not description:
            description = self.generate_description(task_key, run_name, run_id)
        body['description'] = description
        if tags:
            body['tags'] = tags
        artifact = await self.send_post_request(
            self.endpoint_url + '/artifacts', headers, body
        )
        return artifact.get('_id'), path, tags or {}

    async def add_artifact_tags(self, token, artifacts, path, key, value):
        headers = {"x-api-key": token, "Content-Type": "application/json"}
        artifact = artifacts.get(path)
        if not artifact:
            raise ValueError("Artifact doesn't exists."
                             "Use arcee.artifact() to create one")
        artifact['tags'][key] = value
        body = {'tags': artifact['tags']}
        uri = f'{self.endpoint_url}/artifacts/{artifact["id"]}'
        await self.send_patch_request(uri, headers, body)
        return artifact["id"], path, artifact['tags']



import copy


class TestDataAzure:
    @staticmethod
    def get_test_data():
        return {
            "compute": {
                "azEnvironment": "AzurePublicCloud",
                "customData": "",
                "evictionPolicy": "",
                "extendedLocation": {"name": "", "type": ""},
                "isHostCompatibilityLayerVm": "false",
                "licenseType": "",
                "location": "westeurope",
                "name": "vlad-ml-mini",
                "offer": "0001-com-ubuntu-server-focal",
                "osProfile": {
                    "adminUsername": "ubuntu",
                    "computerName": "vlad-ml-mini",
                    "disablePasswordAuthentication": "true",
                },
                "osType": "Linux",
                "placementGroupId": "",
                "plan": {"name": "", "product": "", "publisher": ""},
                "platformFaultDomain": "0",
                "platformSubFaultDomain": "",
                "platformUpdateDomain": "0",
                "priority": "",
                "provider": "Microsoft.Compute",
                "publicKeys": [
                    {
                        "keyData": "ssh-rsa AAAAB\n",
                        "path": "/home/ubuntu/.ssh/authorized_keys",
                    }
                ],
                "publisher": "canonical",
                "resourceGroupName": "vlad-rg",
                "resourceId": "/subscriptions/..",
                "securityProfile": {
                    "secureBootEnabled": "false",
                    "virtualTpmEnabled": "false",
                },
                "sku": "20_04-lts-gen2",
                "storageProfile": {
                    "dataDisks": [],
                    "imageReference": {
                        "id": "",
                        "offer": "0001-com-ubuntu-server-focal",
                        "publisher": "canonical",
                        "sku": "20_04-lts-gen2",
                        "version": "latest",
                    },
                    "osDisk": {
                        "caching": "ReadWrite",
                        "createOption": "FromImage",
                        "diffDiskSettings": {"option": ""},
                        "diskSizeGB": "30",
                        "encryptionSettings": {"enabled": "false"},
                        "image": {"uri": ""},
                        "managedDisk": {
                            "id": "/subscriptions/7ab2fde4b3b6df3300",
                            "storageAccountType": "Premium_LRS",
                        },
                        "name": "vlad-ml-mini_OsDisk_1_c63ece7f1e0340aab0",
                        "osType": "Linux",
                        "vhd": {"uri": ""},
                        "writeAcceleratorEnabled": "false",
                    },
                    "resourceDisk": {"size": "16384"},
                },
                "subscriptionId": "fff",
                "tags": "",
                "tagsList": [],
                "userData": "",
                "version": "20.04.202209050",
                "virtualMachineScaleSet": {"id": ""},
                "vmId": "7940abd0-a8c6-403f-8102-e47738e3cd4a",
                "vmScaleSetName": "",
                "vmSize": "Standard_D2s_v3",
                "zone": "1",
            },
            "network": {
                "interface": [
                    {
                        "ipv4": {
                            "ipAddress": [
                                {
                                    "privateIpAddress": "172.18.0.5",
                                    "publicIpAddress": "",
                                }
                            ],
                            "subnet": [
                                {
                                    "address": "172.18.0.0",
                                    "prefix": "24"
                                }
                            ],
                        },
                        "ipv6": {"ipAddress": []},
                        "macAddress": "000D3A46E876",
                    }
                ]
            },
        }

    @classmethod
    def get_test_data_no_network(cls):
        d = copy.copy(cls.get_test_data())
        d["network"] = dict()
        return d



# flake8: noqa: F401
from .arcee import (init, send, tag, milestone, info, finish, error, stage,
                    dataset, hyperparam, model, model_version,
                    model_version_alias, model_version_tag, artifact,
                    artifact_tag)



import asyncio
import atexit
import time
import threading
import warnings

from kiroframe_arcee.sender.sender import Sender
from kiroframe_arcee.collectors.console import (
    acquire_console, release_console)
from kiroframe_arcee.name_generator import NameGenerator
from kiroframe_arcee.utils import single


class Job(threading.Thread):
    def __init__(self, shutdown_flag, *args, **kwargs):
        # TODO: typing
        threading.Thread.__init__(self)
        self.__shutdown_flag = shutdown_flag
        self.__kw = kwargs

    def s_noblock(self, sender, run, token):
        return asyncio.run(sender.send_proc_data(run, token))

    def job(self):
        args = self.__kw.get("meth_args", list())
        self.s_noblock(*args)

    def run(self):
        sleep = self.__kw.get("sleep")
        if not sleep or not isinstance(sleep, int):
            # 1 second by default
            sleep = 1
        while not self.__shutdown_flag.is_set():
            self.job()
            time.sleep(sleep)


@single
class Arcee:
    def __init__(
        self, token=None, task_key=None, endpoint_url=None, ssl=True
    ):
        self.shutdown_flag = threading.Event()
        self.token = token
        self.task_key = task_key
        self.sender = Sender(endpoint_url, ssl, self.shutdown_flag)
        self.hb = None
        self._run = None
        self._tags = dict()
        self._name = None
        self._hyperparams = dict()
        self._dataset = None
        self._model = None
        self._model_version = None
        self._model_version_tags = dict()
        self._model_version_aliases = list()
        self._artifacts = dict()

    @property
    def run(self):
        return self._run

    @run.setter
    def run(self, value):
        self._run = value

    @property
    def tags(self):
        return self._tags

    @tags.setter
    def tags(self, value):
        k, v = value
        self._tags.update({k: v})

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def hyperparams(self):
        return self._hyperparams

    @hyperparams.setter
    def hyperparams(self, value):
        k, v = value
        self._hyperparams.update({k: v})

    @property
    def dataset(self):
        return self._dataset

    @dataset.setter
    def dataset(self, value):
        self._dataset = value

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            finish()
        else:
            error()
        return exc_type is None

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, value):
        self._model = value

    @property
    def model_version(self):
        return self._model_version

    @model_version.setter
    def model_version(self, value):
        self._model_version = value

    @property
    def model_version_tags(self):
        return self._model_version_tags

    @model_version_tags.setter
    def model_version_tags(self, value):
        k, v = value
        self._model_version_tags.update({k: v})

    @property
    def model_version_aliases(self):
        return self._model_version_aliases

    @model_version_aliases.setter
    def model_version_aliases(self, value):
        aliases = set(self._model_version_aliases)
        aliases.add(value)
        self._model_version_aliases = list(aliases)

    @property
    def artifacts(self):
        return self._artifacts

    @artifacts.setter
    def artifacts(self, value):
        id_, path, tags = value
        self._artifacts[path] = {
            'id': id_,
            'tags': tags
        }


def _unhandled_finish():
    arcee = Arcee()
    if not arcee.shutdown_flag.is_set():
        warnings.warn(
            "Run terminated unexpectedly! Please ensure that you use "
            "`arcee.init()` as a context manager (`with arcee.init():`) or "
            "explicitly call `arcee.finish()` / `arcee.error()`",
            UserWarning
        )
        finish()


def init(
    token, task_key, run_name=None, endpoint_url=None, ssl=True, period=1
):
    acquire_console()
    arcee = Arcee(token, task_key, endpoint_url, ssl)
    name = (
        run_name if run_name is not None else NameGenerator.get_random_name()
    )
    arcee.name = name
    run_id = asyncio.run(arcee.sender.get_run_id(task_key, token, name))["id"]
    arcee.run = run_id
    arcee.hb = Job(
        meth_args=(arcee.sender, run_id, token),
        sleep=period,
        shutdown_flag=arcee.shutdown_flag,
    )
    arcee.hb.start()
    atexit.register(_unhandled_finish)
    asyncio.run(
        arcee.sender.send_stats(
            arcee.token,
            {"project": arcee.task_key, "run": arcee.run, "data": {}},
        )
    )
    return arcee


def hyperparam(key, value):
    """
    Add hyperparameter
    Args:
        key: string
        value: float
    Returns:
    """
    arcee = Arcee()
    arcee.hyperparams = (key, value)
    asyncio.run(arcee.sender.add_hyperparams(
        arcee.run, arcee.token, arcee.hyperparams))


def tag(key, value):
    arcee = Arcee()
    arcee.tags = (key, value)
    asyncio.run(arcee.sender.add_tags(arcee.run, arcee.token, arcee.tags))


def milestone(value):
    arcee = Arcee()
    asyncio.run(arcee.sender.add_milestone(arcee.run, arcee.token, value))


def stage(name):
    arcee = Arcee()
    asyncio.run(arcee.sender.create_stage(arcee.run, arcee.token, name))


def dataset(path, name=None, description=None, labels=None):
    arcee = Arcee()
    if arcee.dataset is None:
        arcee.dataset = path
        asyncio.run(arcee.sender.register_dataset(
            arcee.token, arcee.run, arcee.name, arcee.task_key, path, name,
            description, labels
        ))


def _send_console():
    arcee = Arcee()
    try:
        asyncio.run(
            arcee.sender.send_console(
                arcee.run,
                arcee.token
            )
        )
    except Exception:
        pass


def finish():
    release_console()
    arcee = Arcee()
    _send_console()
    try:
        asyncio.run(
            arcee.sender.change_state(
                arcee.run,
                arcee.token,
                2,
                True,
            )
        )
    finally:
        arcee.shutdown_flag.set()
        arcee.hb.join()


def error():
    release_console()
    arcee = Arcee()
    _send_console()
    try:
        asyncio.run(
            arcee.sender.change_state(
                arcee.run,
                arcee.token,
                3,
                True,
            )
        )
    finally:
        arcee.shutdown_flag.set()
        arcee.hb.join()


def info():
    arcee = Arcee()
    return arcee.__dict__


def send(data):
    arcee = Arcee()
    asyncio.run(
        arcee.sender.send_stats(
            arcee.token,
            {"project": arcee.task_key, "run": arcee.run, "data": data},
        )
    )


def model(key, path=None):
    arcee = Arcee()
    arcee.model = asyncio.run(
        arcee.sender.add_model(
            arcee.token, key
        )
    )
    asyncio.run(
        arcee.sender.create_model_version(
            arcee.run, arcee.model, arcee.token, path=path
        )
    )


def model_version(version):
    arcee = Arcee()
    asyncio.run(
        arcee.sender.add_version(
            arcee.run, arcee.model, arcee.token, version
        )
    )


def model_version_alias(alias):
    arcee = Arcee()
    arcee.model_version_aliases = alias
    asyncio.run(
        arcee.sender.add_version_aliases(
            arcee.run, arcee.model, arcee.token, arcee.model_version_aliases
        )
    )


def model_version_tag(key, value):
    arcee = Arcee()
    arcee.model_version_tags = (key, value)
    asyncio.run(
        arcee.sender.add_version_tags(
            arcee.run, arcee.model, arcee.token, arcee.model_version_tags
        )
    )


def artifact(path, name=None, description=None, tags=None):
    arcee = Arcee()
    arcee.artifacts = asyncio.run(
        arcee.sender.add_artifact(
            arcee.token, arcee.run, arcee.name, arcee.task_key, path, name,
            description, tags
        )
    )


def artifact_tag(path, key, value):
    arcee = Arcee()
    arcee.artifacts = asyncio.run(
        arcee.sender.add_artifact_tags(
            arcee.token, arcee.artifacts, path, key, value
        )
    )



import math
import os
import concurrent.futures
from functools import reduce

import psutil

from kiroframe_arcee.libs.GPUtil import GPUtil
from kiroframe_arcee.utils import run_async


# tune accuracy depending on #cpus
_MEASURE_TIME = 1 - 1 / (os.cpu_count())
_TIME_INTERVALS = (_MEASURE_TIME + 0.05, _MEASURE_TIME + 0.01)


class Collector:
    executor = concurrent.futures.ThreadPoolExecutor(max_workers=10)

    @staticmethod
    def _gpu_stats():
        gpus = GPUtil.getGPUs()
        len_gpus = len(gpus)
        if len_gpus < 1:
            return {}
        avg_gpu_load = reduce(
            lambda x, y: x + y, map(lambda z: z.load, gpus)
        ) / len(gpus)
        avg_gpu_memory_free = reduce(
            lambda x, y: x + y, map(lambda z: z.memoryFree, gpus)
        ) / len(gpus)
        avg_gpu_memory_total = reduce(
            lambda x, y: x + y, map(lambda z: z.memoryTotal, gpus)
        ) / len(gpus)
        avg_gpu_memory_used = reduce(
            lambda x, y: x + y, map(lambda z: z.memoryUsed, gpus)
        ) / len(gpus)
        stats = {
            "avg_gpu_memory_free": avg_gpu_memory_free,
            "avg_gpu_memory_total": avg_gpu_memory_total,
            "avg_gpu_memory_used": avg_gpu_memory_used,
        }
        # get rid of devices with limited support
        if not math.isnan(avg_gpu_load):
            stats["avg_gpu_load"] = avg_gpu_load * 100
        return stats

    @staticmethod
    def _ps_stats(cpu_interval=0.5, proc_interval=0.5):
        load1, load5, load15 = psutil.getloadavg()
        cpu_load = psutil.cpu_percent(interval=cpu_interval, percpu=True)
        cpu_percent = round(sum(cpu_load) / psutil.cpu_count(), 2)
        # get proc pid
        process = psutil.Process(os.getpid())
        # physical mem according to https://psutil.readthedocs.io/en/latest/
        physical_mem = psutil.virtual_memory().total
        swap_mem = psutil.swap_memory().total
        # virtual memory used by process
        proc_vmem = process.memory_info().vms
        # resident state memory used by process
        proc_rss = process.memory_info().rss
        cpu_proc = min(
            [
                round(
                    process.cpu_percent(
                        interval=proc_interval) / psutil.cpu_count(),
                    2,
                ),
                cpu_percent,
            ]
        )

        ps_stats = {
            "cpu_count": psutil.cpu_count(),
            "cpu_percent": cpu_percent,
            "cpu_percent_percpu": cpu_load,
            "load_average": [load1, load5, load15],
            "cpu_usage": (load15 / psutil.cpu_count()) * 100,
            "used_ram_percent": psutil.virtual_memory()[2],
            "used_ram_mb": psutil.virtual_memory()[3] / (1024 * 1024),
        }

        proc_stats = {
            # process cpu usage in % (per core)
            "cpu": cpu_proc,
            "mem": {
                # process virtual memory usage, p - percent, t - value
                "vms": {
                    "p": "{:.3f}".format(
                        proc_vmem / (physical_mem + swap_mem)
                    ),
                    "t": proc_vmem,
                },
                # process resident state memory usage, p - percent, t - value
                "rss": {
                    "p": "{:.3f}".format(proc_rss / physical_mem),
                    "t": proc_rss,
                },
            },
        }
        return ps_stats, proc_stats

    @classmethod
    def _collect_stats(cls):
        cpu_interval, proc_interval = _TIME_INTERVALS
        gpu_stats = cls._gpu_stats()
        ps_stats, ps_info = cls._ps_stats(cpu_interval, proc_interval)

        return {
            "ps_stats": ps_stats,
            "gpu_stats": gpu_stats,
            "proc": ps_info,
        }

    @classmethod
    async def collect_stats(cls):
        return await run_async(cls._collect_stats, executor=cls.executor)



import concurrent.futures
import psutil

from kiroframe_arcee.utils import run_async


class Collector:
    executor = concurrent.futures.ThreadPoolExecutor(max_workers=3)

    @classmethod
    def _collect(cls) -> str:
        return ' '.join(psutil.Process().cmdline())

    @classmethod
    async def collect(cls):
        return await run_async(cls._collect, executor=cls.executor)
